/*
 * Controlador Java de la Secretaria de Estado de Administraciones Publicas
 * para el DNI electronico.
 *
 * El Controlador Java para el DNI electronico es un proveedor de seguridad de JCA/JCE
 * que permite el acceso y uso del DNI electronico en aplicaciones Java de terceros
 * para la realizacion de procesos de autenticacion, firma electronica y validacion
 * de firma. Para ello, se implementan las funcionalidades KeyStore y Signature para
 * el acceso a los certificados y claves del DNI electronico, asi como la realizacion
 * de operaciones criptograficas de firma con el DNI electronico. El Controlador ha
 * sido disenado para su funcionamiento independiente del sistema operativo final.
 *
 * Copyright (C) 2012 Direccion General de Modernizacion Administrativa, Procedimientos
 * e Impulso de la Administracion Electronica
 *
 * Este programa es software libre y utiliza un licenciamiento dual (LGPL 2.1+
 * o EUPL 1.1+), lo cual significa que los usuarios podran elegir bajo cual de las
 * licencias desean utilizar el codigo fuente. Su eleccion debera reflejarse
 * en las aplicaciones que integren o distribuyan el Controlador, ya que determinara
 * su compatibilidad con otros componentes.
 *
 * El Controlador puede ser redistribuido y/o modificado bajo los terminos de la
 * Lesser GNU General Public License publicada por la Free Software Foundation,
 * tanto en la version 2.1 de la Licencia, o en una version posterior.
 *
 * El Controlador puede ser redistribuido y/o modificado bajo los terminos de la
 * European Union Public License publicada por la Comision Europea,
 * tanto en la version 1.1 de la Licencia, o en una version posterior.
 *
 * Deberia recibir una copia de la GNU Lesser General Public License, si aplica, junto
 * con este programa. Si no, consultelo en <http://www.gnu.org/licenses/>.
 *
 * Deberia recibir una copia de la European Union Public License, si aplica, junto
 * con este programa. Si no, consultelo en <http://joinup.ec.europa.eu/software/page/eupl>.
 *
 * Este programa es distribuido con la esperanza de que sea util, pero
 * SIN NINGUNA GARANTIA; incluso sin la garantia implicita de comercializacion
 * o idoneidad para un proposito particular.
 */
package es.gob.jmulticard.card.dnie;

import es.gob.jmulticard.HexUtils;
import es.gob.jmulticard.card.CryptoCard;
import es.gob.jmulticard.card.Location;
import es.gob.jmulticard.card.PrivateKeyReference;

/** Clave privada de un DNIe. La clase no contiene la clave privada en si, sino una referencia a ella
 * y una referencia al propio DNIe, con el canal seguro establecido.
 * @author Tom&aacute;s Garc&iacute;a-Mer&aacute;s. */
public final class DniePrivateKeyReference implements PrivateKeyReference {

	private final CryptoCard dnieCard;
	private final byte[] identifier;
	private final Location keyPath;
	private final String label;
	private transient final byte reference;
	private transient final int keySize;

	/** Crea una referencia a una clave privada del DNIe.
	 * @param card Tarjeta en la que se almacena la clave privada.
	 * @param id Identificador de la clave.
	 * @param path Ruta interna de la clave.
	 * @param lbl Etiqueta de la clave.
	 * @param ref Referencia interna de la clave.
	 * @param keyLength Tama&ntilde;o en bits de la clave privada. */
	public DniePrivateKeyReference(final CryptoCard card,
			                       final byte[] id,
			                       final Location path,
			                       final String lbl,
			                       final byte ref,
			                       final int keyLength) {
		this.dnieCard = card;
		this.identifier = id != null ? id.clone() : null;
		this.keyPath = path;
		this.label = lbl;
		this.reference = ref;
		this.keySize = keyLength;
	}

	@Override
	public String toString() {
		return "Clave privada de DNIe con etiqueta '" + this.label + //$NON-NLS-1$
			"', identificador '" + HexUtils.hexify(this.identifier, false) + //$NON-NLS-1$
				"' y ruta '" + this.keyPath + "'"; //$NON-NLS-1$ //$NON-NLS-2$
	}

	/** Recupera el manejador de la tarjeta en la que se almacena la clave.
	 * @return Manejador de la tarjeta. */
	public CryptoCard getDnieCard() {
		return this.dnieCard;
	}

	/** Recupera el identificador de la clave.
	 * @return Identificador de la clave. */
	public byte[] getIdentifier() {
		return this.identifier;
	}

	/** Recupera la ruta de la clave.
	 * @return Ruta de la clave. */
	public Location getKeyPath() {
		return this.keyPath;
	}

	/** Recupera la etiqueta de la clave.
	 * @return Etiqueta de la clave. */
	public String getLabel() {
		return this.label;
	}

	/** Recupera la referencia de la clave.
	 * @return Referencia de la clave. */
	public byte getKeyReference() {
		return this.reference;
	}

	/** Obtiene el tam&ntilde;o en bits de la clave.
	 * @return Tam&ntilde;o en bits de la clave. */
	public int getKeyBitSize() {
		return this.keySize;
	}

}
